open ExtString;;
open ExtList;;
let (|>) x f = f x;;
let (>>) f g x = g (f x);;
module Bar = Bigarray.Array1;;

module DnaFlat = 
  struct
    type t = { mutable df_arr : (int, Bigarray.int8_unsigned_elt, Bigarray.c_layout) Bar.t; 
               mutable df_cur : int; mutable df_last : int }
      
    let create sz = 
      { df_arr = Bar.create Bigarray.int8_unsigned Bigarray.c_layout sz;
        df_cur = 0; df_last = 0 }
        
    let of_string s = 
      let len = String.length s in
      let d = create len in
      String.enum s |> Enum.iteri (fun i c-> d.df_arr.{i} <- int_of_char c);
      { d with df_last = len }
      
    let length d = d.df_last - d.df_cur
    let current d = if d.df_cur < d.df_last then char_of_int d.df_arr.{d.df_cur} else 'e'  
    let nth d n = 
      let i = d.df_cur + n in if i < d.df_last then char_of_int d.df_arr.{i} else 'e'
    let skip n d = d.df_cur <- d.df_cur + n    
    let copyiter d = { d with df_cur = d.df_cur} (*i*)
    let getnext d =
      if d.df_cur < d.df_last then 
        let c = char_of_int d.df_arr.{d.df_cur} in
        (d.df_cur <- d.df_cur+1; c)
      else 'e'
      
    let append d ch = 
      let sz = Bar.dim d.df_arr in
      if d.df_last >= sz then
       (let na = Bar.create Bigarray.int8_unsigned Bigarray.c_layout (sz*2) in
        Bar.blit d.df_arr (Bar.sub na 0 sz);
        d.df_arr <- na) ;
      d.df_arr.{d.df_last} <- int_of_char ch; 
      d.df_last <- d.df_last + 1
      
    let iter_intp dna pos f =
      if dna.df_cur + pos > dna.df_last then failwith "Dna.iter_intp: pos2big" else
      for i = dna.df_cur + pos to dna.df_last - 1 do f dna.df_arr.{i}  done

    let iter_int dna f = iter_intp dna 0 f 

    let to_string dna = 
      let dcp = copyiter dna in    
      Enum.init (length dcp) (fun i-> char_of_int dcp.df_arr.{dcp.df_cur+i}) |> String.of_enum
          
    let sub dna st en = {dna with df_cur = dna.df_cur + st; df_last = dna.df_cur + en}
    
    let append_dna dna s =
      let sz = Bar.dim dna.df_arr in
      let len = length s in
      if dna.df_last + len > sz then failwith "append_dna no space" else
      Bar.blit (Bar.sub s.df_arr s.df_cur len) (Bar.sub dna.df_arr dna.df_last len);
      dna.df_last <- dna.df_last + len
         
    let of_file fname = 
      let fd = Unix.openfile fname [Unix.O_RDONLY] 0o640 in
      let a = Bar.map_file fd Bigarray.int8_unsigned Bigarray.c_layout false (-1) in
      { df_arr = a; df_cur = 0; df_last = Bar.dim a}
    
  end;;

module Tree = 
  struct
    type node_info = { length : int; nc : int }
    type tree_t = TLeaf of DnaFlat.t * node_info | TNode of tree_t * tree_t * node_info  
    
    let create sz = 
      let d = DnaFlat.create sz in TLeaf(d, {length = 0; nc = 1})
        
    let of_string s =     
      let d = DnaFlat.of_string s in TLeaf(d, {length = DnaFlat.length d; nc = 1})
      
    let length = function TLeaf(d, ni) -> ni.length | TNode(t1,t2,ni) -> ni.length
    let cnt = function TLeaf(d, ni) -> ni.nc | TNode(t1,t2,ni) -> ni.nc
    
    let rec append t ch = 
      match t with
      | TLeaf(d, ni) -> DnaFlat.append d ch; TLeaf(d, {ni with length = ni.length + 1})
      | TNode(t1,t2,ni) -> failwith "Tree.append should not be called for big trees"
    
    let append_dna t1 t2 =       
      TNode(t1, t2, {length = length t1 + length t2; nc = cnt t1 + cnt t2 + 1})
        
    let rec sub t st en =
      match t with
      | TLeaf(d,ni) -> 
          let len = DnaFlat.length d in
          if en > len then failwith "Tree.sub on leaf: en too big" else
          TLeaf(DnaFlat.sub d st en, {length = en-st; nc = 1})
      | TNode(t1,t2,ni) -> 
        let len1 = length t1 and len2 = length t2 in        
        if en > len1+len2 then failwith "Tree.sub on node: en too big" else
        if en <= len1 then sub t1 st en else
        if st >= len1 then sub t2 (st-len1) (en-len1) else
        let st1 = sub t1 st len1 and st2 = sub t2 0 (en-len1) in
        let c1 = cnt st1 and c2 = cnt st2 in
        TNode(st1, st2, {length = en-st; nc = c1+c2+1})
      
    let rec segment t pos =
      match t with
      | TLeaf(d,ni)-> 
        if pos > DnaFlat.length d then failwith "segment.leaf: pos2big" else (d, pos)
      | TNode(t1,t2,ni)-> 
        let len1 = length t1 in if pos < len1 then segment t1 pos else segment t2 (pos-len1)
         
    let iter_int t pos f =
      let rec elist pos lst = function
        | TNode(t1,t2,ni)-> let len1 = length t1 in 
            if pos < len1 then elist pos (elist 0 lst t2) t1 else elist (pos-len1) lst t2              
        | TLeaf(d,ni)-> (d, pos)::lst
      in elist pos [] t |> List.iter (fun (d,pos)-> DnaFlat.iter_intp d pos f)    
    
    let of_file fname = 
      let d = DnaFlat.of_file fname in TLeaf(d, {length = DnaFlat.length d; nc = 1})
      
    let flatten t = 
      let d = DnaFlat.create (length t) in
      let rec dlist lst = function
        | TNode(t1,t2,ni)-> dlist (dlist lst t2) t1 
        | TLeaf(d,ni)-> d::lst
      in List.iter (DnaFlat.append_dna d) (dlist [] t);
      TLeaf(d, {length = DnaFlat.length d; nc = 1})
      
  end;;

module DnaTree = 
  struct
    type t = { mutable tr : Tree.tree_t; mutable pos : int; 
               mutable curseg : DnaFlat.t; mutable offs : int;
               mutable buf : Tree.tree_t option;  }
      
    let create sz = 
      let t = Tree.create sz in  let (seg, ofs) = Tree.segment t 0 in
      { tr = t; pos = 0; curseg = seg; offs = ofs; buf = None }
      
    let of_string s = 
      let d = create (String.length s) in
      let t = String.enum s |> Enum.fold (fun c t -> Tree.append t c) d.tr in
      { d with tr = t }
      
    let length d = Tree.length d.tr - d.pos
    
    let current d = 
      let i = d.pos - d.offs in
      if i < DnaFlat.length d.curseg then DnaFlat.nth d.curseg i else 
      try let (seg, ofs) = Tree.segment d.tr d.pos in
        d.curseg <- seg; d.offs<- d.pos - ofs; DnaFlat.nth d.curseg ofs
      with Failure _-> 'e'  
    
    let nth d n = 
      try let (seg, ofs) = Tree.segment d.tr (n + d.pos) in DnaFlat.nth seg ofs
      with Failure _-> 'e' 
    
    let skip n d = d.pos <- d.pos + n    
    
    let getnext d =
      let c = current d in d.pos <- d.pos + 1; c
    
    let enum d = (*i*)
      let pos = d.pos in
      Enum.init (length d) (fun i-> let (seg, ofs) = Tree.segment d.tr (i + pos) in DnaFlat.nth seg ofs )
      
    let to_string = enum >> String.of_enum
      
    let commit d =
      match d.buf with None->() | Some bt-> d.tr <- Tree.append_dna d.tr bt; d.buf <- None
          
    let iter_int d f =
      Tree.iter_int d.tr d.pos f
    
    let sub dna st en = 
      let t = Tree.sub dna.tr (st+dna.pos) (en+dna.pos) in let (seg, ofs) = Tree.segment t 0 in
      { tr = t; pos = 0; curseg = seg; offs = ofs; buf = None }
       
    let append_dna dna s =
      let lt = match dna.buf with None-> dna.tr | Some bt-> Tree.append_dna dna.tr bt in
      let t = Tree.append_dna lt (Tree.sub s.tr s.pos (Tree.length s.tr)) in
      dna.tr <- t; dna.buf <- None
      
    let append d ch =
      let tt =  match d.buf with Some bt -> bt | None -> Tree.create 500 in
      d.buf <- Some( Tree.append tt ch)

    let of_file fname =
      let t = Tree.of_file fname in let (seg, ofs) = Tree.segment t 0 in
      { tr = t; pos = 0; curseg = seg; offs = ofs; buf = None }
      
    exception Found of int
      
    let search dna i s =
      let len = length s and dlen = length dna in
      if len > (dlen - i) then None else
      if len = 0 then Some (i) else
      let buf = Array.make len 0 and sa = Array.make len 0 and len1 = len-1 in
      let j = ref 0 in iter_int s (fun x-> sa.(!j) <- x; incr j);     
      let rec cmploop j k =        
        if k >= len then true else
        if buf.(j mod len)=sa.(k) then cmploop (j+1) (k+1) else false
      in
      try
        let di = ref 0 in
        Tree.iter_int dna.tr (dna.pos+i) (fun x->
          buf.(!di mod len)<-x;
          if !di >= len1 && cmploop (!di+1) 0 then raise (Found !di);
          incr di);                
        None
      with Found di-> Some (di+1+i)
    
    let ndnalen x = 1
    let compact d = 
      if Tree.cnt d.tr > 400 then
      let t = Tree.flatten d.tr in let (seg, ofs) = Tree.segment t d.pos in
      (d.tr <- t; d.curseg <- seg; d.offs <- d.pos - ofs)
      
    let cnt d = Tree.cnt d.tr       
  end;;
